#
#  Copyright (c) 1998, 1999, Digital Creations, Fredericksburg, VA, USA.  
#  All rights reserved.
#  
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are
#  met:
#  
#    o Redistributions of source code must retain the above copyright
#      notice, this list of conditions, and the disclaimer that follows.
#  
#    o Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions, and the following disclaimer in
#      the documentation and/or other materials provided with the
#      distribution.
#  
#    o Neither the name of Digital Creations nor the names of its
#      contributors may be used to endorse or promote products derived
#      from this software without specific prior written permission.
#  
#  
#  THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS AND CONTRIBUTORS *AS
#  IS* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
#  TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
#  PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DIGITAL
#  CREATIONS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
#  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
#  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
#  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
#  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
#  USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
#  DAMAGE.
#
#  $Id: ociProc.py,v 1.5 1999/01/14 17:44:56 jim Exp $
#
#  If you have questions regarding this software,
#  contact:
# 
#    Digital Creations L.C.  
#    info@digicool.com
# 
#    (540) 371-6909
#
##############################################################################
from string import join, lower
from oci_ import oparse
import sys, oci_
from ociBind import bindScalar, _bindBufferScalar
from ociUtil import RowID, oDate, error
from Buffer import Buffer

if not hasattr(sys, 'exc_info'):
    def exc_info():
        return sys.exc_type, sys.exc_value, sys.exc_traceback
    sys.exc_info=exc_info

class Procedure:
    "Oracle stored procedures"

    def __init__(self, c, fname, desc):
        self._c=c
        self.__name__=fname

        # Rearrange description by argument:
        desc=apply(map,(None,)+desc)
        desc.sort()
        
        # Split data into separate signatures:
        last=-1
        sigs=[]
        sig=[]
        seen_default=0
        for d in desc:
            (overload, pos, level, name, type, has_default, iomode,
             dtsize, prec, scale, radix) = d
            if level != 0: raise error, (
                "Procedures with record and array arguments aren't supported")

            if overload != last:
                sig=[]
                sigs.append(sig)
                last=overload
                seen_default=0

            if has_default:
                if iomode: raise error, (
                    "Huh, how can an output argument, %s (%d of %d), "
                    "have a default?" % (name, pos, overload))
                seen_default=1
            elif seen_default: raise error, (
                "Argument without default follows argument with default")
                
            sig.append(d)

        if not sigs: sigs.append([])

        for i in range(len(sigs)):
            sigs[i]=compile_sig(fname, sigs[i])

        self._sigs=sigs

    def _error(self):
        c=self._c
        rc=oci_.cda_def_rc_get(oci_.cucda(c))
        raise error, (rc, oci_.OracleErrorMessage(c.lda, rc))

    def __call__(self, *args, **kw):
        sigs=self._sigs
        d=None
        oc=self._c
        r=[]
        cruft=[]
        out=[]
        iocursors=[]
        try:
            for sig in self._sigs:
                try:
                    sql, inp, oup, iop = apply(sig,args,kw)
                    if oparse(oc, sql, 0, 1): self._error()
                    for (name, type, size, prec, scale, radix), v in inp:
                        cruft.append(bindScalar(oc, name, v))
                    for oupi in oup:
                        name, type, size, prec, scale, radix = oupi
                        bo=bindOutput(oc, name, type, size, prec, scale)
                        cruft.append(bo)
                        out.append(bo)
                    for (name, type, size, prec, scale, radix), v in iop:
                        cruft.append(bindScalar(oc, name, v, 1))
                        if type==oci_.SQLT_CUR and hasattr(v,'_describe'):
                            iocursors.append(v)
                    
                    if oci_.oexec(oc): self._error()
                    for c in iocursors: c._describe()
                    if not out: return None
                    r=[]
                    for arcode, name, v, alen, ind, f, size in out:
                        if ind[0] < 0: v=None
                        else:
                            if f==1: v=v[:alen[0]]
                            else:
                                v=v[0]
                                if f is not None: v=f(v)
                        r.append(v)
                    if len(r)==1: return r[0]
                except: t,v,tb=sys.exc_info()
            raise t, v, tb
        finally: tb=None

        
_sql_default_val=[]

def bindOutput(c, name, type, size, prec, scale,
               types={1:'s', 2:'d', 11:11, 12:7, 23:'s', 96:'s', },
               funcs={1:1, 2:None, 11:RowID, 12:oDate, 23:1, 96:1, },
               ):

    try: bt=types[type]
    except: raise error, 'Unsupported output type, %s, for %s' % (type, name)
    f=funcs[type]
    
    if type==2:
        if scale:
            type=4
        else:
            type=3
            bt='i'
        
    arcode=Buffer(1,'H')
    alen=Buffer(1,'H')
    ind=Buffer(1,'h')

    if bt=='s':
        if size==0:
            if type==1: size=2000
            elif type==96 or type==23: size=255
            else: raise error, 'Odd type/size combination %s/%s' % (type, size)
        v=Buffer(size)
    else:
        v=Buffer(1,bt)
        size=v.itemsize
        
    alen[0]=size
    oci_.bind(c, name, v, size, type, ind, alen, arcode, 0, None)
    return arcode, name, v, alen, ind, f, size

def compile_sig(fname, sig):
    args_by_mode=[[],[],[]]
    argnames_by_mode=[[],None,[]]
    src=[]
    sql=[fname+'(']
    delim=':'
    ifs=[]
    _sig=[]
    globals={'_d': _sql_default_val, '__s':_sig}
    i=0
    for d in sig:
        (overload, pos, level, name, type, has_default, mode,
         dtsize, prec, scale, radix) = d
        if has_default:
            # Hm, we need to generate an if test.
            # if name is _d: return 'sql',
            s=join(sql,'')+')'
            if s[-2:]=='()': s=s[:-2]
            s="begin\n  %s;\nend;\n" % s
            ifs.append('  if %s is _d: return %s, (%s), (%s), (%s)'
                       % (name, `s`,
                          args_by_mode[0] and
                          join(map(lambda ades, aname:
                                   "(_s[%d], %s)" % (ades, aname),
                                   args_by_mode[0], argnames_by_mode[0]),
                               ', ')+',' or '',
                          args_by_mode[1] and
                          join(map(lambda ades:
                                   "_s[%d]" % ades,
                                   args_by_mode[1]),
                               ', ')+',' or '',
                          args_by_mode[2] and
                          join(map(lambda ades, aname:
                                   "(_s[%d], %s)" % (ades, aname),
                                   args_by_mode[2], argnames_by_mode[2]),
                               ', ')+',' or '',
                          ))
        if not name:
            # Function return
            name='ocidb_function_return__'
            sql.insert(0, ':ocidb_function_return__ := ')
        else:
            sql.append(delim+name)
            delim=',:'
        _sig.append((name, type, dtsize, prec, scale, radix))
        args_by_mode[mode].append(i)
        if mode != 1:
            if has_default: src.append('%s=_d' % name)
            else: src.append(name)
            argnames_by_mode[mode].append(name)
        i=i+1

    # Add default case, all arguments:
    s=join(sql,'')+')'
    if s[-2:]=='()': s=s[:-2]
    s="begin\n  %s;\nend;\n" % s
    ifs.append('  return %s, (%s), (%s), (%s)'
               % (`s`,
                  args_by_mode[0] and
                  join(map(lambda ades, aname:
                           "(_s[%d], %s)" % (ades, aname),
                           args_by_mode[0], argnames_by_mode[0]),
                       ', ')+',' or '',
                  args_by_mode[1] and
                  join(map(lambda ades:
                           "_s[%d]" % ades,
                           args_by_mode[1]),
                       ', ')+',' or '',
                  args_by_mode[2] and
                  join(map(lambda ades, aname:
                           "(_s[%d], %s)" % (ades, aname),
                           args_by_mode[2], argnames_by_mode[2]),
                       ', ')+',' or '',
                  ))

    # OK, we have everything we need.  Gnerate the code.
    src=("def sig(%s):\n  _s=__s\n%s\n" %
         (join(src,","), join(ifs,"\n")))

    # print src

    exec src in globals
    sig=globals['sig']
    del globals['sig'] # Break circular link, for the heck of it
    return sig

class StoredProcedureNamespace:
    # Provide access to a database's stored procedure namespaces
    def __init__(self, c, package=''):
        self._c=c
        self._p=package

    def __getattr__(self, name):
        # Try to create a stored-procedure wrapper.  If we can't,
        # Then assume we have a package.

        # Prepend package name:
        if self._p: name='%s.%s' % (self._p, name)

        # First, try to get a procedure desc:
        c=self._c._c
        try:
            pdesc=oci_.odessp(c.lda, name)
            # OK, got this far, must be a procedure, not a package.
            r=Procedure(c, name, pdesc)
            setattr(self, name, r)
            return r
        except: pass
        
        # Hm, that didn'r work, let's see if it's a package:
        c=self._c
        try:
            c.execute(
                "select count(*) from all_objects"
                " where lower(object_name)=:name"
                " and object_type='PACKAGE'",
                name=lower(name))
            r=c.fetchone()
            if r and r[0]:
                r=StoredProcedureNamespace(c, name)
                setattr(self, name, r)
                return r
        except: pass

        # Darn.  Maybe it's a schema name.
        try:
            c.execute(
                "select count(*) from all_users"
                " where lower(username)=:name",
                name=lower(name))
            r=c.fetchone()
            if r and r[0]:
                r=StoredProcedureNamespace(c, name)
                setattr(self, name, r)
                return r
        except: pass
            
        raise error, 'no usable procedure named '+name
                

