#
#  Copyright (c) 1998, 1999, Digital Creations, Fredericksburg, VA, USA.  
#  All rights reserved.
#  
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are
#  met:
#  
#    o Redistributions of source code must retain the above copyright
#      notice, this list of conditions, and the disclaimer that follows.
#  
#    o Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions, and the following disclaimer in
#      the documentation and/or other materials provided with the
#      distribution.
#  
#    o Neither the name of Digital Creations nor the names of its
#      contributors may be used to endorse or promote products derived
#      from this software without specific prior written permission.
#  
#  
#  THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS AND CONTRIBUTORS *AS
#  IS* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
#  TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
#  PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DIGITAL
#  CREATIONS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
#  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
#  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
#  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
#  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
#  USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
#  DAMAGE.
#
#  $Id: ocidb.py,v 1.10 1999/08/27 18:12:10 petrilli Exp $
#
#  If you have questions regarding this software,
#  contact:
# 
#    Digital Creations L.C.  
#    info@digicool.com
# 
#    (540) 371-6909
#
##############################################################################

"""DBI-Compliant Oracle Database Interface
"""

import oci_, ociCurs, ociProc
try: import oci_8
except: oci_8=None

from ociUtil import error
from string import join, split

# If threads are enabled, then put Oracle in thread-safe mode:
try: import thread
except: thread=None
if thread is not None: oci_.opinit(oci_.OCI_EV_TSF)
del thread

class ocidb:
    "Oracle Database Connections"

    if oci_8 is None:
        def __init__(self, connection_string):
            "Open a database connection"
            self.__dict__['_d']=oci_.Connect(connection_string)
    else:
        def __init__(self, connection_string):
            "Open a database connection"
            # Waaa, O8 wants a different format. Waaa
            up, s = (split(connection_string, '@')+[''])[:2]
            u,  p = (split(up, '/')+[''])[:2]
            sc=oci_8.Logon(u,p,s)
            self.__dict__['_d']=sc.lda()

    def close(self):
        oci_.ologof(self._d)

    def commit(self):
        oci_.ocom(self._d)

    def rollback(self):
        oci_.orol(self._d)

    def cursor(self):
        return ociCurs.Cursor(self._d)

    def prepare(self, statement):
        c = ociCurs.Handle(self._d)
        c._parse(statement)
        return c

    def callproc(self, name, params=()):
        return apply(getattr(self.procedures, name),params)

    def objects(self, all=0):
        if all:
            t='ALL'
            e=', OWNER'
        else:
            t='USER'
            e=''
        c=self._c
        c.execute(
            'select OBJECT_NAME, OBJECT_TYPE%s from %s_OBJECTS' % (e,t)
            )
        return c.fetchall()

    def getSource(self, proc):
        c=self._c
        c.execute(
            'select line, text from all_source where name=:p1', (proc,)
            )
        r=[]
        a=r.append
        for line, text in c.fetchall(): a((line,text))
        r.sort()
        return join(map(lambda i: i[1], r),'')
        

    # Blech! The DBI says:
    #
    # all Cursor Object attributes and methods
    # 
    # For databases that do not have cursors and for simple applications
    # that do not require the complexity of a cursor, a Connection Object
    # should respond to each of the attributes and methods of the Cursor
    # Object.  Databases that have cursor can implement this by using an
    # implicit, internal cursor.
    def __getattr__(self, name):
        d=self.__dict__
        if name=='procedures':
            c=self.cursor() # Allocate a cursor for procs
            d[name]=p=c.procedures
            return p
        if d.has_key('_c'): c=d['_c']
        else: d['_c']=c=self.cursor()
        if name=='_c': return c
        if hasattr(c,name): return getattr(c,name)
        raise AttributeError, name

    def __setattr__(self, name, v):
        d=self.__dict__
        if name=='arraysize': self._c.arraysize=v
        else:
            raise AttributeError, 'Unassignable attribute, '+name

        
Connect=ocidb
